/******************************************************************************
* Buzzer control driver - v0.3
* 1. Control system buzzer with specific pattern.
*
* Author: Simon Chang, UG.
*
* Change history:
* v0.1 Initial version.
* v0.2 Add ioctl function.
* v0.3 Modify set frequency as real frequency.
******************************************************************************/
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <asm/delay.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/cdev.h>
#include <linux/timer.h>
#include <asm/param.h>


#define DEVNO			127			//device number
#define	DEVNAME			"kiri_bz"	//device name

#define TIMER_PORT		0x43
#define COUNT_PORT		0x42
#define SPK_PORT			0x61
#define SYS_OCS_TICK	1193182

#define TIMER_DELAY		HZ

#define BZ_PATTERN_CONTINUOUS	0
#define BZ_PATTERN_ON_OFF			1
#define BZ_OPERATION_OFF			0
#define BZ_OPERATION_ON				1

#define IOCTL_GET_BZ_OPERATION	0
#define IOCTL_SET_BZ_OPERATION	1
#define IOCTL_GET_BZ_PATTERN		6
#define IOCTL_SET_BZ_PATTERN		3
#define IOCTL_GET_BZ_FREQ				4
#define IOCTL_SET_BZ_FREQ				7

int bz_major = DEVNO;
char *bz_devname = DEVNAME;

#if defined(CONFIG_BUFFALO_PLATFORM)
#include "kiri_common.h"
struct class *kiri_bz_cls = NULL;
#else
extern struct class *kiri_cls;
#endif
struct device *kiri_bz_dev;
unsigned int bz_freq = 262;		// default frequency is "Do"
int bz_pattern = BZ_PATTERN_CONTINUOUS;
int bz_operation = BZ_OPERATION_OFF;
static struct timer_list bz_timer;



/*******************************************************************************
* function: timer_function
* argument:
* retvalue:
* description:
*******************************************************************************/
static void timer_function(unsigned long data)
{
	if (data == BZ_OPERATION_ON) {
		outb(inb(SPK_PORT) & 0xfc, SPK_PORT);				// turn off buzzer
		bz_timer.data = BZ_OPERATION_OFF;
	}
	else {
		outb(inb(SPK_PORT) | 3, SPK_PORT);				// turn on buzzer
		bz_timer.data = BZ_OPERATION_ON;
	}
	mod_timer(&bz_timer, jiffies + TIMER_DELAY);
}

/*******************************************************************************
* function: bz_ioctl
* argument:
* retvalue:
* description:
*******************************************************************************/
int bz_ioctl(struct file *file,unsigned int cmd, unsigned long para)
{
	int ret = 0;
	int temp_pattern;

	switch (cmd) {
		case IOCTL_GET_BZ_OPERATION:
			ret = bz_operation;
			if (bz_operation) printk("on\n");
			else printk("off\n");
		break;

		case IOCTL_SET_BZ_OPERATION:
			if(para == BZ_OPERATION_ON) {
				if (bz_operation == BZ_OPERATION_ON) return 0;
				bz_operation = BZ_OPERATION_ON;
			}
			else if(para == BZ_OPERATION_OFF) {
				if (bz_operation == BZ_OPERATION_OFF) return 0;
				bz_operation = BZ_OPERATION_OFF;
			}
			else
				goto IOCTL_WRONG_CMD;

			if (bz_operation) {		// turn on
				if (bz_pattern) {
 					init_timer(&bz_timer);
					bz_timer.expires = jiffies + TIMER_DELAY;
					bz_timer.data = BZ_OPERATION_ON;
					bz_timer.function = timer_function;
					add_timer(&bz_timer);
				}
				outb(inb(SPK_PORT) | 3, SPK_PORT);				// turn on buzzer
			}
			else {		// turn off
				if (bz_pattern) del_timer(&bz_timer);
				outb(inb(SPK_PORT) & 0xfc, SPK_PORT);				// turn off buzzer
			}
		break;

		case IOCTL_GET_BZ_PATTERN:
			ret = bz_pattern;
			printk("  ret=%d\n", bz_pattern);
			if (bz_pattern) printk("on-off\n");
			else printk("continuous\n");
		break;

		case IOCTL_SET_BZ_PATTERN:
			if(para == BZ_PATTERN_ON_OFF)
				temp_pattern = BZ_PATTERN_ON_OFF;
			else if(para == BZ_PATTERN_CONTINUOUS)
				temp_pattern = BZ_PATTERN_CONTINUOUS;
			else goto IOCTL_WRONG_CMD;

			if (temp_pattern == bz_pattern) return 0;

			if (bz_operation == BZ_OPERATION_ON) {
				if (temp_pattern == BZ_PATTERN_ON_OFF) {
 					init_timer(&bz_timer);
					bz_timer.expires = jiffies + TIMER_DELAY;
					bz_timer.data = BZ_OPERATION_ON;
					bz_timer.function = timer_function;
					add_timer(&bz_timer);
				}
				else {
					del_timer(&bz_timer);
				}
				outb(inb(SPK_PORT) | 3, SPK_PORT);				// turn on buzzer
			}
			bz_pattern = temp_pattern;
		break;

		case IOCTL_GET_BZ_FREQ:
			ret = bz_freq;
			printk("%d\n", bz_freq);
		break;

		case IOCTL_SET_BZ_FREQ:
			bz_freq = (unsigned int)para;
			if (bz_freq == 0) goto IOCTL_WRONG_CMD;
			outb(0xb6, TIMER_PORT);
			outb((unsigned char)(SYS_OCS_TICK/bz_freq), COUNT_PORT);			/* write LSB */
			outb((unsigned char)((SYS_OCS_TICK/bz_freq) >> 8), COUNT_PORT);		/* write MSB */
		break;

		default:
			goto IOCTL_WRONG_CMD;
		break;
	}
	return ret;

IOCTL_WRONG_CMD:
	printk("wrong command\n");
	return ret;
}

struct file_operations bz_fops = {
	//open: bz_open,
	unlocked_ioctl: bz_ioctl,
	//release: bz_close,
	owner: THIS_MODULE,
};

/*******************************************************************************
* function: show_bz_freq
* function: store_bz_freq
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_bz_freq(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", bz_freq);
}
static ssize_t store_bz_freq(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned long temp_freq;

	temp_freq = simple_strtoul(buf, NULL, 10);
	if (temp_freq == 0) return -EINVAL;
	bz_freq = (unsigned int)temp_freq;
	temp_freq = (u16)(SYS_OCS_TICK / bz_freq);
	outb(0xb6, TIMER_PORT);
	outb((unsigned char)temp_freq, COUNT_PORT);				/* write LSB */
	outb((unsigned char)(temp_freq >> 8), COUNT_PORT);		/* write MSB */

	return count;
}
DEVICE_ATTR(freq, 0644, show_bz_freq, store_bz_freq);

/*******************************************************************************
* function: show_bz_pattern
* function: store_bz_pattern
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_bz_pattern(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (bz_pattern) return sprintf(buf, "on-off\n");
	else return sprintf(buf, "continuous\n");

}
static ssize_t store_bz_pattern(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	int temp_pattern;

	if((strncmp(buf, "on-off", strlen("on-off")) == 0))
		temp_pattern = BZ_PATTERN_ON_OFF;
	else if((strncmp(buf, "continuous", strlen("continuous")) == 0))
		temp_pattern = BZ_PATTERN_CONTINUOUS;

	if (temp_pattern == bz_pattern) return count;

	if (bz_operation == BZ_OPERATION_ON) {
		if (temp_pattern == BZ_PATTERN_ON_OFF) {
 			init_timer(&bz_timer);
			bz_timer.expires = jiffies + TIMER_DELAY;
			bz_timer.data = BZ_OPERATION_ON;
			bz_timer.function = timer_function;
			add_timer(&bz_timer);
		}
		else {
			del_timer(&bz_timer);
		}
		outb(inb(SPK_PORT) | 3, SPK_PORT);				// turn on buzzer
	}

	bz_pattern = temp_pattern;

	return count;
}
DEVICE_ATTR(pattern, 0644, show_bz_pattern, store_bz_pattern);

/*******************************************************************************
* function: show_bz_operation
* function: store_bz_operation
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_bz_operation(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (bz_operation) return sprintf(buf, "on\n");
	else return sprintf(buf, "off\n");

}
static ssize_t store_bz_operation(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	if((strncmp(buf, "on", strlen("on")) == 0)) {
		if (bz_operation == BZ_OPERATION_ON) return count;
		bz_operation = BZ_OPERATION_ON;
	}
	else if((strncmp(buf, "off", strlen("off")) == 0)) {
		if (bz_operation == BZ_OPERATION_OFF) return count;
		bz_operation = BZ_OPERATION_OFF;
	}
	else
		return count;

	if (bz_operation) {		// turn on
		if (bz_pattern) {
 			init_timer(&bz_timer);
			bz_timer.expires = jiffies + TIMER_DELAY;
			bz_timer.data = BZ_OPERATION_ON;
			bz_timer.function = timer_function;
			add_timer(&bz_timer);
		}
		outb(inb(SPK_PORT) | 3, SPK_PORT);				// turn on buzzer
	}
	else {		// turn off
		if (bz_pattern) del_timer(&bz_timer);
		outb(inb(SPK_PORT) & 0xfc, SPK_PORT);				// turn off buzzer
	}

	return count;
}
DEVICE_ATTR(operation, 0644, show_bz_operation, store_bz_operation);



/*******************************************************************************
* function: kiri_bz_ctrl_init
* argument:
* retvalue:
* description:
*******************************************************************************/
static int __init kiri_bz_ctrl_init(void)
{
	int ret;

	printk(KERN_ALERT "buzzer control module init...\n");

	/* Register device, and may be allocating major# */
	ret = register_chrdev(bz_major, bz_devname, &bz_fops);
	if (ret < 0) {
		printk("%s cannot register\n", bz_devname);
		return(ret);
	}
	else {
		if (bz_major == 0) bz_major = ret;
	}

#if defined(CONFIG_BUFFALO_PLATFORM)
	kiri_bz_cls = KiriClassGet();
#endif
	if (kiri_bz_cls)
	{
		kiri_bz_dev = device_create(kiri_bz_cls, NULL, MKDEV(bz_major, 0), NULL, "bz");
		ret = device_create_file(kiri_bz_dev, &dev_attr_freq);
		ret = device_create_file(kiri_bz_dev, &dev_attr_pattern);
		ret = device_create_file(kiri_bz_dev, &dev_attr_operation);
	}

	// set default frequency
	outb(0xb6, TIMER_PORT);
	outb((unsigned char)(SYS_OCS_TICK/bz_freq), COUNT_PORT);			/* write LSB */
	outb((unsigned char)((SYS_OCS_TICK/bz_freq) >> 8), COUNT_PORT);		/* write MSB */

	return 0;
}

/*******************************************************************************
* function: kiri_bz_ctrl_exit_cleanup
* argument:
* retvalue:
* description:
*******************************************************************************/
static void __exit kiri_bz_ctrl_exit_cleanup(void)
{
	//int ret;

	printk(KERN_ALERT "buzzer control module exit...\n");

	outb(inb(SPK_PORT) & 0xfc, SPK_PORT);				// turn off buzzer

	device_remove_file(kiri_bz_dev, &dev_attr_freq);
	device_remove_file(kiri_bz_dev, &dev_attr_pattern);
	device_remove_file(kiri_bz_dev, &dev_attr_operation);
	device_destroy(kiri_bz_cls, MKDEV(bz_major, 0));
	unregister_chrdev(bz_major, bz_devname);
#if defined(CONFIG_BUFFALO_PLATFORM)
	KiriClassDestroy();
#endif
}

module_init(kiri_bz_ctrl_init);
module_exit(kiri_bz_ctrl_exit_cleanup);
MODULE_LICENSE("GPL");
