/*
 * HDD control driver - 0.3
 * 1. Detect if HDD is presented or not.
 * 2. Get the power status of HDD.
 * 3. Set the power status of HDD.
 * 4. Get/Set Lederr.
 *
 * Author: Simon Chang, USI.
 *
 *
 * History:
 * v0.1: Initial version for 12 HDDs.
 * v0.2: Change string "matsu" to "kiri"
 * v0.3  Add Lederr function.
 */

#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <asm/uaccess.h>
#include <linux/i2c.h>
#include <linux/mutex.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/version.h>


#define HDD_CTRL_MODULE_VERSION "0.1"
#define HDD_CTRL_MODULE_NAME "hdd control module"


#define MAX_SUPPORTED_DISKS	12

/* Define for procfs directory and file */
#if defined(CONFIG_BUFFALO_PLATFORM)
#include <buffalo/kernevnt.h>
#define MODULE_DIR		"buffalo"
#define MODULE_GPIO_DIR		MODULE_DIR"/gpio"
#define HDD_PRESENT_DIR		MODULE_GPIO_DIR"/hotplug"
#define HDD_POWER_DIR		MODULE_GPIO_DIR"/power_control"
#else
#define MODULE_DIR		"kiri"
#define MODULE_GPIO_DIR		"kiri/gpio"
#define HDD_PRESENT_DIR		"kiri/gpio/hotplug"
#define HDD_POWER_DIR		"kiri/gpio/power_control"
#endif

/* Define for PCA9505 GPIO expander */
#define GPI_PORT0		0x00
#define GPI_PORT1		0x01
#define GPI_PORT2		0x02
#define GPI_PORT3		0x03
#define GPI_PORT4		0x04

#define GPO_PORT0		0x08
#define GPO_PORT1		0x09
#define GPO_PORT2		0x0a
#define GPO_PORT3   0x0b
#define GPO_PORT4   0x0c

/* Define for LED Command */
#define LED_CMD_ERR		0
#define LED_CMD_ON		1
#define LED_CMD_OFF		2
#define LED_CMD_BLINK	3

#define HDD_PRESENT0_BIT	(1 << 0)
#define HDD_PRESENT1_BIT	(1 << 1)
#define HDD_PRESENT2_BIT	(1 << 2)
#define HDD_PRESENT3_BIT	(1 << 3)
#define HDD_PRESENT4_BIT	(1 << 4)
#define HDD_PRESENT5_BIT	(1 << 5)
#define HDD_PRESENT6_BIT	(1 << 6)
#define HDD_PRESENT7_BIT	(1 << 7)
#define HDD_PRESENT8_BIT	(1 << 0)
#define HDD_PRESENT9_BIT	(1 << 1)
#define HDD_PRESENT10_BIT	(1 << 2)
#define HDD_PRESENT11_BIT	(1 << 3)

#define HDD_PRESENT0_MASK	(~HDD_PRESENT0_BIT)
#define HDD_PRESENT1_MASK	(~HDD_PRESENT1_BIT)
#define HDD_PRESENT2_MASK	(~HDD_PRESENT2_BIT)
#define HDD_PRESENT3_MASK	(~HDD_PRESENT3_BIT)
#define HDD_PRESENT4_MASK	(~HDD_PRESENT4_BIT)
#define HDD_PRESENT5_MASK	(~HDD_PRESENT5_BIT)
#define HDD_PRESENT6_MASK	(~HDD_PRESENT6_BIT)
#define HDD_PRESENT7_MASK	(~HDD_PRESENT7_BIT)
#define HDD_PRESENT8_MASK	(~HDD_PRESENT8_BIT)
#define HDD_PRESENT9_MASK	(~HDD_PRESENT9_BIT)
#define HDD_PRESENT10_MASK	(~HDD_PRESENT10_BIT)
#define HDD_PRESENT11_MASK	(~HDD_PRESENT11_BIT)

#define HDD_POWER0_BIT		(1 << 4)
#define HDD_POWER1_BIT		(1 << 5)
#define HDD_POWER2_BIT		(1 << 6)
#define HDD_POWER3_BIT		(1 << 7)
#define HDD_POWER4_BIT		(1 << 0)
#define HDD_POWER5_BIT		(1 << 1)
#define HDD_POWER6_BIT		(1 << 2)
#define HDD_POWER7_BIT		(1 << 3)
#define HDD_POWER8_BIT		(1 << 4)
#define HDD_POWER9_BIT		(1 << 5)
#define HDD_POWER10_BIT		(1 << 6)
#define HDD_POWER11_BIT		(1 << 7)

#define HDD_POWER0_MASK		(~HDD_POWER0_BIT)
#define HDD_POWER1_MASK		(~HDD_POWER1_BIT)
#define HDD_POWER2_MASK		(~HDD_POWER2_BIT)
#define HDD_POWER3_MASK		(~HDD_POWER3_BIT)
#define HDD_POWER4_MASK		(~HDD_POWER4_BIT)
#define HDD_POWER5_MASK		(~HDD_POWER5_BIT)
#define HDD_POWER6_MASK		(~HDD_POWER6_BIT)
#define HDD_POWER7_MASK		(~HDD_POWER7_BIT)
#define HDD_POWER8_MASK		(~HDD_POWER8_BIT)
#define HDD_POWER9_MASK		(~HDD_POWER9_BIT)
#define HDD_POWER10_MASK	(~HDD_POWER10_BIT)
#define HDD_POWER11_MASK	(~HDD_POWER11_BIT)

#define HDD_ERRLED0_BIT		(1 << 0)
#define HDD_ERRLED1_BIT		(1 << 1)
#define HDD_ERRLED2_BIT		(1 << 2)
#define HDD_ERRLED3_BIT		(1 << 3)
#define HDD_ERRLED4_BIT		(1 << 4)
#define HDD_ERRLED5_BIT		(1 << 5)
#define HDD_ERRLED6_BIT		(1 << 6)
#define HDD_ERRLED7_BIT		(1 << 7)
#define HDD_ERRLED8_BIT		(1 << 0)
#define HDD_ERRLED9_BIT		(1 << 1)
#define HDD_ERRLED10_BIT	(1 << 2)
#define HDD_ERRLED11_BIT	(1 << 3)

#define HDD_ERRLED0_MASK	(~HDD_ERRLED0_BIT)
#define HDD_ERRLED1_MASK	(~HDD_ERRLED1_BIT)
#define HDD_ERRLED2_MASK	(~HDD_ERRLED2_BIT)
#define HDD_ERRLED3_MASK	(~HDD_ERRLED3_BIT)
#define HDD_ERRLED4_MASK	(~HDD_ERRLED4_BIT)
#define HDD_ERRLED5_MASK	(~HDD_ERRLED5_BIT)
#define HDD_ERRLED6_MASK	(~HDD_ERRLED6_BIT)
#define HDD_ERRLED7_MASK	(~HDD_ERRLED7_BIT)
#define HDD_ERRLED8_MASK	(~HDD_ERRLED8_BIT)
#define HDD_ERRLED9_MASK	(~HDD_ERRLED9_BIT)
#define HDD_ERRLED10_MASK	(~HDD_ERRLED10_BIT)
#define HDD_ERRLED11_MASK	(~HDD_ERRLED11_BIT)

struct kiri_hdd_info_st {
	u8 hdd_gpio_present_port;
	u8 hdd_present_bit;
	u8 hdd_gpio_power_port;
	u8 hdd_power_bit;
	u8 hdd_power_mask;
};

static struct kiri_hdd_info_st kiri_hdd_info[] = {
	{GPI_PORT0,	HDD_PRESENT0_BIT,	GPO_PORT1,	HDD_POWER0_BIT,		HDD_POWER0_MASK},
	{GPI_PORT0,	HDD_PRESENT1_BIT,	GPO_PORT1,	HDD_POWER1_BIT,		HDD_POWER1_MASK},
	{GPI_PORT0,	HDD_PRESENT2_BIT,	GPO_PORT1,	HDD_POWER2_BIT,		HDD_POWER2_MASK},
	{GPI_PORT0,	HDD_PRESENT3_BIT,	GPO_PORT1,	HDD_POWER3_BIT,		HDD_POWER3_MASK},
	{GPI_PORT0,	HDD_PRESENT4_BIT,	GPO_PORT2,	HDD_POWER4_BIT,		HDD_POWER4_MASK},
	{GPI_PORT0,	HDD_PRESENT5_BIT,	GPO_PORT2,	HDD_POWER5_BIT,		HDD_POWER5_MASK},
	{GPI_PORT0,	HDD_PRESENT6_BIT,	GPO_PORT2,	HDD_POWER6_BIT,		HDD_POWER6_MASK},
	{GPI_PORT0,	HDD_PRESENT7_BIT,	GPO_PORT2,	HDD_POWER7_BIT,		HDD_POWER7_MASK},
	{GPI_PORT1,	HDD_PRESENT8_BIT,	GPO_PORT2,	HDD_POWER8_BIT,		HDD_POWER8_MASK},
	{GPI_PORT1,	HDD_PRESENT9_BIT,	GPO_PORT2,	HDD_POWER9_BIT,		HDD_POWER9_MASK},
	{GPI_PORT1,	HDD_PRESENT10_BIT,	GPO_PORT2,	HDD_POWER10_BIT,	HDD_POWER10_MASK},
	{GPI_PORT1,	HDD_PRESENT11_BIT,	GPO_PORT2,	HDD_POWER11_BIT,	HDD_POWER11_MASK},
};

/* Global variables of proc filesystem */
static struct proc_dir_entry *hdd_power_dir;
static struct proc_dir_entry *hdd_present_dir;

/* Global variables of sysfs filesystem */
#if defined(CONFIG_BUFFALO_PLATFORM)
#include "kiri_common.h"
struct class *kiri_hdd_cls;
#else
extern struct class *kiri_cls;
#endif
extern struct device *kiri_led_dev;

#define led_major		126

/* for hotplug */

typedef enum _sata_plug_state {
	SATA_STAT_UNKNOWN,
	SATA_STAT_PLUGGED,
	SATA_STAT_UNPLUGGED,
}SATA_PLUG_STATE;

struct sata_hotplug_data_st {
	SATA_PLUG_STATE presentpinstat;
	SATA_PLUG_STATE prevplugstat;
	unsigned int loops;
};

/* for polling timer */
#define SATA_POL_INTERVAL       HZ * 3
#define SATA_POL_LOOPS          2

#define PLUGGED_EVENT_MSG       "SATA %d plugged"
#define UNPLUGGED_EVENT_MSG     "SATA %d unplugged"

struct timer_list sata_hotplug_polling_timer;
struct sata_hotplug_data_st sata_hotplug_data[MAX_SUPPORTED_DISKS];

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,39)
spinlock_t kiri_hdd_status_lock = SPIN_LOCK_UNLOCKED;
#else
spinlock_t kiri_hdd_status_lock;
#endif
static struct workqueue_struct *interrupt_work_wq;
static struct work_struct interrupt_work;


#define PCA9505_SLAVE_ADDR	0x20

/* Addresses to scan */
static const unsigned short hdd_gpio_i2c[] = { PCA9505_SLAVE_ADDR, I2C_CLIENT_END };
struct i2c_client *kiri_gpio_exp_client;
struct kiri_hdd_data {
	struct mutex update_lock;
	u8 gpio_port_data;
};

/*******************************************************************************
* function: read_hdd_present_status
* argument:
* retvalue:
*******************************************************************************/
static int read_hdd_present_status(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	struct kiri_hdd_info_st *hdd_info = (struct kiri_hdd_info_st *)data;
	struct kiri_hdd_data *gpio_data = i2c_get_clientdata(kiri_gpio_exp_client);

	if (off > 0) {
		*eof = 1;
		return 0;
	}

	mutex_lock(&gpio_data->update_lock);
	gpio_data->gpio_port_data = (u8)i2c_smbus_read_byte_data(kiri_gpio_exp_client, hdd_info->hdd_gpio_present_port);
	mutex_unlock(&gpio_data->update_lock);

	if (gpio_data->gpio_port_data & hdd_info->hdd_present_bit)
		len = sprintf(page, "unplugged\n");
	else
		len = sprintf(page, "plugged\n");
	*eof = 1;
	return len;
}

/*******************************************************************************
* function: read_hdd_power_status
* argument:
* retvalue:
*******************************************************************************/
static int read_hdd_power_status(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	struct kiri_hdd_info_st *hdd_info = (struct kiri_hdd_info_st *)data;
	struct kiri_hdd_data *gpio_data = i2c_get_clientdata(kiri_gpio_exp_client);

	if (off > 0) {
		*eof = 1;
		return 0;
	}

	mutex_lock(&gpio_data->update_lock);
	gpio_data->gpio_port_data = (u8)i2c_smbus_read_byte_data(kiri_gpio_exp_client, hdd_info->hdd_gpio_power_port);
	mutex_unlock(&gpio_data->update_lock);

	if (gpio_data->gpio_port_data & hdd_info->hdd_power_bit)
		len = sprintf(page, "on\n");
	else
		len = sprintf(page, "off\n");

	*eof = 1;
	return len;

}

/*******************************************************************************
* function: set_hdd_power
* argument:
* retvalue:
*******************************************************************************/
static int set_hdd_power(struct file *file, const char *buffer, unsigned long count, void *data)
{
	struct kiri_hdd_info_st *hdd_info = (struct kiri_hdd_info_st *)data;
	struct kiri_hdd_data *gpio_data = i2c_get_clientdata(kiri_gpio_exp_client);

	mutex_lock(&gpio_data->update_lock);
	gpio_data->gpio_port_data = (u8)i2c_smbus_read_byte_data(kiri_gpio_exp_client, hdd_info->hdd_gpio_power_port);
	mutex_unlock(&gpio_data->update_lock);

	if (strncmp(buffer, "off", 3) == 0) {
		if (gpio_data->gpio_port_data & hdd_info->hdd_power_bit) {
			gpio_data->gpio_port_data &= hdd_info->hdd_power_mask;
			mutex_lock(&gpio_data->update_lock);
			i2c_smbus_write_byte_data(kiri_gpio_exp_client, hdd_info->hdd_gpio_power_port, gpio_data->gpio_port_data);
			mutex_unlock(&gpio_data->update_lock);
		}
	}
	else if (strncmp(buffer, "on", 2) == 0) {
		if ((gpio_data->gpio_port_data & hdd_info->hdd_power_bit) != hdd_info->hdd_power_bit) {
			gpio_data->gpio_port_data |= hdd_info->hdd_power_bit;
			mutex_lock(&gpio_data->update_lock);
			i2c_smbus_write_byte_data(kiri_gpio_exp_client, hdd_info->hdd_gpio_power_port, gpio_data->gpio_port_data);
			mutex_unlock(&gpio_data->update_lock);
		}
	}
	else {
		return -EINVAL;
	}

	return count;
}

/*******************************************************************************
* function: set_hdd_lederr_on
* argument:
* retvalue:
* description:
*******************************************************************************/
void set_hdd_lederr_on(int hddno)
{
	struct kiri_hdd_data *gpio_data = i2c_get_clientdata(kiri_gpio_exp_client);
	u8 temp = 0x01;
	u8 port;

	if (hddno < 8) {
		temp <<= hddno;
		port = GPO_PORT3;
	}
	else if (hddno < 12) {
		hddno -= 8;
		temp <<= hddno;
		port = GPO_PORT4;
	}
	else return;

	mutex_lock(&gpio_data->update_lock);
	gpio_data->gpio_port_data = (u8)i2c_smbus_read_byte_data(kiri_gpio_exp_client, port);
	gpio_data->gpio_port_data &= ~temp;
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, port, gpio_data->gpio_port_data);
	mutex_unlock(&gpio_data->update_lock);
}

/*******************************************************************************
* function: set_hdd_lederr_off
* argument:
* retvalue:
* description:
*******************************************************************************/
void set_hdd_lederr_off(int hddno)
{
	struct kiri_hdd_data *gpio_data = i2c_get_clientdata(kiri_gpio_exp_client);
	u8 temp = 0x01;
	u8 port;

	if (hddno < 8) {
		temp <<= hddno;
		port = GPO_PORT3;
	}
	else if (hddno < 12) {
		hddno -= 8;
		temp <<= hddno;
		port = GPO_PORT4;
	}
	else return;

	mutex_lock(&gpio_data->update_lock);
	gpio_data->gpio_port_data = (u8)i2c_smbus_read_byte_data(kiri_gpio_exp_client, port);
	gpio_data->gpio_port_data |= temp;
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, port, gpio_data->gpio_port_data);
	mutex_unlock(&gpio_data->update_lock);
}

/*******************************************************************************
* function: get_hdd_lederr_status
* argument:
* retvalue:
* description:
*******************************************************************************/
int get_hdd_lederr_status(int hddno)
{
	struct kiri_hdd_data *gpio_data = i2c_get_clientdata(kiri_gpio_exp_client);
	u8 temp = 0x01;
	u8 port;

	if (hddno < 8) {
		temp <<= hddno;
		port = GPO_PORT3;
	}
	else if (hddno < 12) {
		hddno -= 8;
		temp <<= hddno;
		port = GPO_PORT4;
	}
	else return -EINVAL;

	mutex_lock(&gpio_data->update_lock);
	gpio_data->gpio_port_data = (u8)i2c_smbus_read_byte_data(kiri_gpio_exp_client, port);
	mutex_unlock(&gpio_data->update_lock);

	return (gpio_data->gpio_port_data & temp);
}

/*******************************************************************************
* function: check_led_cmd_param
* argument:
* retvalue:
* description:
*******************************************************************************/
static int check_led_cmd_param(char *buf, size_t count)
{
	if (strncmp(buf, "on", 2) == 0) {
		if (count == 2) {
			return LED_CMD_ON;
		}
		else if (count == 3) {
			if (buf[2] == '\n') return LED_CMD_ON;
			else return LED_CMD_ERR;
		}
		else return LED_CMD_ERR;
	}
	else if (strncmp(buf, "off", 3) == 0) {
		if (count == 3) {
			return LED_CMD_OFF;
		}
		else if (count == 4) {
			if (buf[3] == '\n') return LED_CMD_OFF;
			else return LED_CMD_ERR;
		}
		else return LED_CMD_ERR;
	}
	else if (strncmp(buf, "blink", 5) == 0) {
		if (count == 5) {
			return LED_CMD_BLINK;
		}
		else if (count == 6) {
			if (buf[5] == '\n') return LED_CMD_BLINK;
			else return LED_CMD_ERR;
		}
		else return LED_CMD_ERR;
	}
	else return LED_CMD_ERR;

}

/*******************************************************************************
* function: show_Lederr0
* function: store_Lederr0
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr0(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(0))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr0(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(0);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(0);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr0, 0644, show_Lederr0, store_Lederr0);

/*******************************************************************************
* function: show_Lederr1
* function: store_Lederr1
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr1(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(1))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr1(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(1);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(1);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr1, 0644, show_Lederr1, store_Lederr1);

/*******************************************************************************
* function: show_Lederr2
* function: store_Lederr2
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr2(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(2))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr2(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(2);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(2);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr2, 0644, show_Lederr2, store_Lederr2);

/*******************************************************************************
* function: show_Lederr3
* function: store_Lederr3
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr3(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(3))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr3(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(3);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(3);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr3, 0644, show_Lederr3, store_Lederr3);

/*******************************************************************************
* function: show_Lederr4
* function: store_Lederr4
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr4(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(4))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr4(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(4);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(4);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr4, 0644, show_Lederr4, store_Lederr4);

/*******************************************************************************
* function: show_Lederr5
* function: store_Lederr5
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr5(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(5))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr5(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(5);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(5);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr5, 0644, show_Lederr5, store_Lederr5);

/*******************************************************************************
* function: show_Lederr6
* function: store_Lederr6
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr6(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(6))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr6(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(6);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(6);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr6, 0644, show_Lederr6, store_Lederr6);

/*******************************************************************************
* function: show_Lederr7
* function: store_Lederr7
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr7(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(7))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr7(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(7);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(7);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr7, 0644, show_Lederr7, store_Lederr7);

/*******************************************************************************
* function: show_Lederr8
* function: store_Lederr8
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr8(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(8))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr8(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(8);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(8);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr8, 0644, show_Lederr8, store_Lederr8);

/*******************************************************************************
* function: show_Lederr9
* function: store_Lederr9
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr9(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(9))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr9(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(9);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(9);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr9, 0644, show_Lederr9, store_Lederr9);

/*******************************************************************************
* function: show_Lederr10
* function: store_Lederr10
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr10(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(10))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr10(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(10);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(10);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr10, 0644, show_Lederr10, store_Lederr10);

/*******************************************************************************
* function: show_Lederr11
* function: store_Lederr11
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_Lederr11(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(11))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr11(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(11);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(11);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr11, 0644, show_Lederr11, store_Lederr11);

/*******************************************************************************
* function: kiri_gpio_exp_detect
* argument:
* retvalue:
* description:
*******************************************************************************/
static int kiri_gpio_exp_detect(struct i2c_client *client, struct i2c_board_info *info)
{
	strlcpy(info->type, "kiri_hdd_gpio", I2C_NAME_SIZE);

	return 0;
}

/*******************************************************************************
* function: kiri_gpio_exp_probe
* argument:
* retvalue:
* description:
*******************************************************************************/
static int kiri_gpio_exp_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	struct kiri_hdd_data *data;
	if (!(data = kzalloc(sizeof(struct kiri_hdd_data), GFP_KERNEL))) {
		return -ENOMEM;
	}

	kiri_gpio_exp_client = client;
	data->gpio_port_data = 0;
	i2c_set_clientdata(kiri_gpio_exp_client, data);
	mutex_init(&data->update_lock);

	// for init PCA9505, because BIOS is not ready.
	struct kiri_hdd_data *gpio_data = i2c_get_clientdata(kiri_gpio_exp_client);
	mutex_lock(&gpio_data->update_lock);
	/* hdd control setting */
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, 0x18, 0xff);
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, 0x19, 0x0f);
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, 0x1a, 0x00);
	/* interrupt setting of sata hotplug status */
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, 0x20, 0xff);
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, 0x21, 0x0f);
	/* lederror setting */
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, 0x1b, 0x00);
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, 0x1c, 0x00);
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, 0x0b, 0xff);
	i2c_smbus_write_byte_data(kiri_gpio_exp_client, 0x0c, 0xff);
	
	mutex_unlock(&gpio_data->update_lock);



	return 0;
}

/*******************************************************************************
* function: kiri_gpio_exp_remove
* argument:
* retvalue:
* description:
*******************************************************************************/
static int kiri_gpio_exp_remove(struct i2c_client *client)
{
	kfree(i2c_get_clientdata(client));
	return 0;
}

static const struct i2c_device_id kiri_gpio_exp_id[] = {
	{ "kiri_hdd_gpio", 0 },
	{ }
};

static struct i2c_driver kiri_gpio_exp_driver = {
	.driver = {
		.name	= "kiri_hdd_gpio",
	},
	.probe		= kiri_gpio_exp_probe,
	.remove		= kiri_gpio_exp_remove,
	.id_table	= kiri_gpio_exp_id,
	.class		= I2C_CLASS_SPD,
	.detect		= kiri_gpio_exp_detect,
	.address_list	= hdd_gpio_i2c,
};

#if defined(CONFIG_BUFFALO_PLATFORM)
#else
/*******************************************************************************
* function: KiriClassGet
* argument:
* retvalue:
* description:
*******************************************************************************/
struct class *KiriClassGet(void)
{
	struct class *tmp_cls;

	if(kiri_cls)
		return kiri_cls;

	tmp_cls = class_create(THIS_MODULE, "kiri");
	if(IS_ERR(tmp_cls)) {
		printk("%s : class_create failed.", __FUNCTION__);
		return NULL;
	}

	kiri_cls = tmp_cls;
	return kiri_cls;
}
#endif

/*******************************************************************************
* function: init_proc_files
* argument:
* retvalue:
* description: create procfs for hdd control:
/ --- proc --- kiri --- gpio -+- hotplug -+- sata0
                              |           +- sata1
                              |           +- sata2
                              |           +- sata3
                              |           +- sata4
                              |           +- sata5
                              |           +- sata6
                              |           +- sata7
                              |           +- sata8
                              |           +- sata9
                              |           +- sata10
                              |           +- sata11
                              |
                              +- power_control -+- hdd0
                                                +- hdd1
                                                +- hdd2
                                                +- hdd3
                                                +- hdd4
                                                +- hdd5
                                                +- hdd6
                                                +- hdd7
                                                +- hdd8
                                                +- hdd9
                                                +- hdd10
                                                +- hdd11
*******************************************************************************/
static int init_proc_files(void)
{
	int retval = 0;
	int i = 0;
	char buf[32];

	/* declare for file and directory of procfs */
	// add process of create "kiri" directory only, because new kernel do not accept
	// path name "kiri/gpio".
	static struct proc_dir_entry *hdd_ctrl_dir;
	static struct proc_dir_entry *hdd_ctrl_gpio_dir;
	static struct proc_dir_entry *hdd_present_status_file[MAX_SUPPORTED_DISKS];
	static struct proc_dir_entry *hdd_power_status_file[MAX_SUPPORTED_DISKS];

	printk("  %s enter...\n", __func__);

#if !defined(CONFIG_BUFFALO_PLATFORM)
	// Create /proc/kiri directory
	hdd_ctrl_dir = proc_mkdir(MODULE_DIR, NULL);
	if(hdd_ctrl_dir == NULL) {
		printk(KERN_ALERT "/proc/kiri directory create fail.....\n");
		goto MODULE_DIR_FAIL;
	}
	printk(KERN_ALERT "/proc/kiri directory create successful.....\n");
#endif

	// Create /proc/kiri/gpio directory
	hdd_ctrl_gpio_dir = proc_mkdir(MODULE_GPIO_DIR, NULL);
	if(hdd_ctrl_gpio_dir == NULL) {
		printk(KERN_ALERT "/proc/kiri/gpio directory create fail.....\n");
		goto MODULE_GPIO_DIR_FAIL;
	}
	printk(KERN_ALERT "/proc/kiri/gpio directory create successful.....\n");

	// Create /proc/kiri/gpio/hotplug directory
	hdd_present_dir = proc_mkdir(HDD_PRESENT_DIR, NULL);
	if(hdd_present_dir == NULL) {
		printk(KERN_ALERT "hotplug directory create fail.....\n");
		goto PRESENT_DIR_FAIL;
	}
	printk(KERN_ALERT "hotplug directory create successful.....\n");

	for(i = 0; i < MAX_SUPPORTED_DISKS; i++)
	{
		sprintf(buf, "sata%d", i);
		hdd_present_status_file[i] = create_proc_read_entry(buf, 0, hdd_present_dir, read_hdd_present_status, NULL);
		if(hdd_present_status_file[i] == NULL)
		{
			printk(KERN_ALERT "HDD%d_PRESENT_STATUS_FILE create fail.....\n", i);
			goto HDD_PRESENT_STATUS_FILE_FAIL;
		}
		printk(KERN_ALERT "HDD%d_PRESENT_STATUS_FILE create successful.....\n", i);
		hdd_present_status_file[i]->data = (void *)&kiri_hdd_info[i];
	}

	// Create /proc/kiri/gpio/power_control directory
	hdd_power_dir = proc_mkdir(HDD_POWER_DIR, NULL);
	if(hdd_power_dir == NULL) {
		printk(KERN_ALERT "power_control directory create fail.....\n");
		goto HDD_POWER_DIR_FAIL;
	}
	printk(KERN_ALERT "power_control directory create successful.....\n");

	for(i = 0; i < MAX_SUPPORTED_DISKS; i++)
	{
		sprintf(buf, "hdd%d", i);
		hdd_power_status_file[i] = create_proc_entry(buf, 0644, hdd_power_dir);
		if(hdd_power_status_file[i] == NULL)
		{
			printk(KERN_ALERT "HDD%d_POWER_STATUS_FILE create fail.....\n", i);
			goto HDD_POWER_STATUS_FILE_FAIL;
		}
		printk(KERN_ALERT "HDD%d_POWER_STATUS_FILE create successful.....\n", i);
		hdd_power_status_file[i]->read_proc = &read_hdd_power_status;
		hdd_power_status_file[i]->write_proc = &set_hdd_power;
		hdd_power_status_file[i]->data = (void *)&kiri_hdd_info[i];
	}

	goto INIT_OK;

HDD_POWER_STATUS_FILE_FAIL:
	for (i = 0; i < MAX_SUPPORTED_DISKS; i++)
	{
		sprintf(buf, "hdd%d", i);
		remove_proc_entry(buf, hdd_power_dir);
	}
	remove_proc_entry(HDD_POWER_DIR, NULL);

HDD_POWER_DIR_FAIL:
HDD_PRESENT_STATUS_FILE_FAIL:
	for (i = 0; i < MAX_SUPPORTED_DISKS; i++)
	{
		sprintf(buf, "sata%d", i);
		remove_proc_entry(buf, hdd_present_dir);
	}

PRESENT_DIR_FAIL:
	remove_proc_entry(HDD_PRESENT_DIR, NULL);

MODULE_GPIO_DIR_FAIL:
	remove_proc_entry(MODULE_GPIO_DIR, NULL);

MODULE_DIR_FAIL:
#if !defined(CONFIG_BUFFALO_PLATFORM)
	remove_proc_entry(MODULE_DIR, NULL);
#endif
	retval = -ENOMEM;

INIT_OK:
	return retval;
}

/* -------------------------------------polling of hotplug status------------------------------------- */

static void SataHotplugPollingUpdatePinstat(void)
{
	struct kiri_hdd_data *gpio_data = i2c_get_clientdata(kiri_gpio_exp_client);
	unsigned int i = 0;
	
	for(i = 0; i < MAX_SUPPORTED_DISKS; i++)
	{
		mutex_lock(&gpio_data->update_lock);
		sata_hotplug_data[i].presentpinstat = (i2c_smbus_read_byte_data(kiri_gpio_exp_client,kiri_hdd_info[i].hdd_gpio_present_port) & kiri_hdd_info[i].hdd_present_bit)? SATA_STAT_UNPLUGGED:SATA_STAT_PLUGGED;
		mutex_unlock(&gpio_data->update_lock);
	}

}

static void SataHotplugPolling(void)
{
	//printk("%s : entered\n", __FUNCTION__);
	unsigned int i = 0;
	char buf[100] = { 0 };
	int ret;

	//Get hotplug status from i2c
	//mdelay(50);
	//spin_lock(&kiri_hdd_status_lock);
	ret = queue_work(interrupt_work_wq, &interrupt_work);
	//spin_unlock(&kiri_hdd_status_lock);
	
	for(i = 0; i < MAX_SUPPORTED_DISKS; i++)
	{
		//printk("sata_hotplug_data[%d].prevplugstat=%d : sata_hotplug_data[%d].presentpinstat=%d\n", i, sata_hotplug_data[i].prevplugstat, i, sata_hotplug_data[i].presentpinstat);
		if(sata_hotplug_data[i].prevplugstat == sata_hotplug_data[i].presentpinstat)
		{
			sata_hotplug_data[i].loops = SATA_POL_LOOPS;
			continue;
		}
		--sata_hotplug_data[i].loops;
		//printk("sata_hotplug_data[%d].loops=%d\n", i, sata_hotplug_data[i].loops);

		if(sata_hotplug_data[i].loops == 0)
		{
			if(sata_hotplug_data[i].presentpinstat == SATA_STAT_PLUGGED){
				snprintf(buf, sizeof(buf), PLUGGED_EVENT_MSG, i);
				printk("%s\n", buf);
			}
			else{
				snprintf(buf, sizeof(buf), UNPLUGGED_EVENT_MSG, i);
				printk("%s\n", buf);
			}
#if defined(CONFIG_BUFFALO_PLATFORM)
			buffalo_kernevnt_queuein(buf);
#endif
			sata_hotplug_data[i].loops = SATA_POL_LOOPS;
			sata_hotplug_data[i].prevplugstat = sata_hotplug_data[i].presentpinstat;
		}
	}
	sata_hotplug_polling_timer.expires += SATA_POL_INTERVAL;
	add_timer(&sata_hotplug_polling_timer);
}

static unsigned long
SataHotplugPollingStart(void)
{
	unsigned int i = 0;
	for(i = 0; i < MAX_SUPPORTED_DISKS; i++)
	{
		sata_hotplug_data[i].prevplugstat = SATA_STAT_UNKNOWN;
		sata_hotplug_data[i].loops = SATA_POL_LOOPS;
	}

        init_timer(&sata_hotplug_polling_timer);
        sata_hotplug_polling_timer.expires = jiffies + SATA_POL_INTERVAL;
        sata_hotplug_polling_timer.function = SataHotplugPolling;
        sata_hotplug_polling_timer.data = NULL;
        add_timer(&sata_hotplug_polling_timer);

        return 0;
}

/*******************************************************************************
* function: hdd_ctrl_init
* argument:
* retvalue:
*******************************************************************************/
static int __init hdd_ctrl_init(void)
{
	int ret;
	
	printk(KERN_ALERT "[%s init proc]\n", HDD_CTRL_MODULE_NAME);

	if (i2c_add_driver(&kiri_gpio_exp_driver)) return -ENODEV;

	/* create files within proc filesystem */
	if (init_proc_files()) return -ENOMEM;

	/* create files within sysfs filesystem */
	kiri_hdd_cls = KiriClassGet();
	if(kiri_hdd_cls){
		if(kiri_led_dev == NULL){
			kiri_led_dev = device_create(kiri_hdd_cls, NULL, MKDEV(led_major, 0), NULL, "led");
		}

		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr0);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr1);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr2);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr3);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr4);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr5);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr6);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr7);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr8);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr9);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr10);
		ret = device_create_file(kiri_led_dev, &dev_attr_Lederr11);

	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39)
	spin_lock_init(&kiri_hdd_status_lock);
#endif
	interrupt_work_wq = create_workqueue("interrupt_queue");
	if(interrupt_work_wq){
  	INIT_WORK(&interrupt_work, SataHotplugPollingUpdatePinstat);
  	SataHotplugPollingStart();
  	
  	printk(KERN_INFO "%s %s initialized...\n",HDD_CTRL_MODULE_NAME, HDD_CTRL_MODULE_VERSION);
  	return 0;
  }
	
	return -1;
}
/*******************************************************************************
* function: hdd_ctrl_exit
* argument:
* retvalue:
*******************************************************************************/
static void __exit hdd_ctrl_exit(void)
{
	int i = 0;
	char buf[32];
	
	
	device_remove_file(kiri_led_dev, &dev_attr_Lederr0);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr1);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr2);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr3);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr4);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr5);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr6);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr7);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr8);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr9);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr10);
	device_remove_file(kiri_led_dev, &dev_attr_Lederr11);
	
#if defined(CONFIG_BUFFALO_PLATFORM)
	KiriClassDestroy();
#endif
	
	printk(KERN_ALERT "[%s exit proc]\n", HDD_CTRL_MODULE_NAME);
	for (i = 0; i < MAX_SUPPORTED_DISKS; i++)
	{
		sprintf(buf, "hdd%d", i);
		remove_proc_entry(buf, hdd_power_dir);
	}
	remove_proc_entry(HDD_POWER_DIR, NULL);
	for (i = 0; i < MAX_SUPPORTED_DISKS; i++)
	{
		sprintf(buf, "sata%d", i);
		remove_proc_entry(buf, hdd_present_dir);
	}
	remove_proc_entry(HDD_PRESENT_DIR, NULL);
	remove_proc_entry(HDD_POWER_DIR, NULL);
	remove_proc_entry(MODULE_GPIO_DIR, NULL);
#if !defined(CONFIG_BUFFALO_PLATFORM)
	remove_proc_entry(MODULE_DIR, NULL);
#endif
	
	flush_workqueue(interrupt_work_wq);
	destroy_workqueue(interrupt_work_wq);
	del_timer(&sata_hotplug_polling_timer);
	
	i2c_del_driver(&kiri_gpio_exp_driver);
	printk(KERN_ALERT "hdd ctrl exit successfully....\n");
}

module_init (hdd_ctrl_init);
module_exit (hdd_ctrl_exit);

MODULE_LICENSE("GPL");
